from __future__ import annotations

import asyncio
import logging
import multiprocessing as mp
from functools import partial
from logging import DEBUG, LogRecord, getLogger
from logging.handlers import QueueHandler
from multiprocessing.context import BaseContext
from queue import Queue
from typing import Callable, Optional

from .func import to_thread

__all__ = ['MultiprocessingLogging']


def example_func():
    '''Used in doctest, defined here to be picklable.'''
    logger = logging.getLogger(__name__)
    logger.warning('warning from another process')


class MultiprocessingLogging:
    '''Collect logging from other processes in the main process.

    Example:

    A function to be executed in another process is example_func(), which is defined
    outside of the docstring because it must be picklable.

    Define the main asynchronous function, in which MultiprocessingLogging is used.

    >>> async def main():
    ...     from concurrent.futures import ProcessPoolExecutor
    ...
    ...     # Start MultiprocessingLogging.
    ...     async with MultiprocessingLogging() as mp_logging:
    ...
    ...         # The initializer is given to ProcessPoolExecutor.
    ...         with ProcessPoolExecutor(initializer=mp_logging.initializer) as executor:
    ...
    ...             # In another process, execute example_func(), which logs a warning.
    ...             future = executor.submit(example_func)
    ...
    ...             # Wait until the function returns.
    ...             future.result()

    When the main() is executed, the warning in example_func() is received in the
    main process.

    To confirm that it works in this example, we add a queue handler to the
    current logger.

    >>> queue = Queue()
    >>> handler = QueueHandler(queue)
    >>> logger = logging.getLogger(__name__)
    >>> logger.addHandler(handler)

    Run the main function.

    >>> asyncio.run(main())

    Check the loggings in the queue.

    >>> log_record = queue.get()
    >>> log_record.getMessage()
    'warning from another process'

    Remove the queue handler added for the example.

    >>> logger.removeHandler(handler)


    Reference: "Logging to a single file from multiple processes" (Logging Cookbook)
    https://docs.python.org/3/howto/logging-cookbook.html#logging-to-a-single-file-from-multiple-processes

    '''

    def __init__(self, mp_context: Optional[BaseContext] = None) -> None:
        mp_context = mp_context or mp.get_context()
        self._q: Queue[LogRecord | None] = mp_context.Queue()
        self._initializer = partial(_initializer, self._q)
        self._task: asyncio.Task | None = None

    @property
    def initializer(self) -> Callable[[], None]:
        '''A callable with no args to be given to ProcessPoolExecutor as initializer.'''
        return self._initializer

    async def open(self):
        self._task = asyncio.create_task(self._listen())

    async def _listen(self) -> None:
        '''Receive loggings from other processes and handle them in the main process.'''
        while (record := await to_thread(self._q.get)) is not None:
            logger = getLogger(record.name)
            if logger.getEffectiveLevel() <= record.levelno:
                logger.handle(record)

    async def close(self) -> None:
        if self._task:
            await to_thread(self._q.put, None)
            await self._task
            self._task = None

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        del exc_type, exc_value, traceback
        await self.close()


def _initializer(queue: Queue[LogRecord]) -> None:
    '''An initializer of ProcessPoolExecutor.'''
    handler = QueueHandler(queue)
    logger = getLogger()
    logger.setLevel(DEBUG)
    logger.addHandler(handler)



from __future__ import annotations

import sys
import threading
from contextlib import contextmanager
from typing import TYPE_CHECKING, Optional

from nextline.process.trace.wrap import FilterByModuleName

if TYPE_CHECKING:
    from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401


@contextmanager
def sys_trace(trace_func: TraceFunc, thread: Optional[bool] = True):
    '''Trace callables in the context and all threads created during the context.

    Notes
    -----
    All new threads created during the context are traced regardless of whether
    the threads are created by code in the context. If the thread option is
    false, no new threads will be traced.

    Example:

    >>> def trace_func(frame, event, arg):
    ...     # will be skipped in the doctest because of TraceSkipModule
    ...     return trace_func

    >>> def callable():
    ...     pass

    >>> with sys_trace(trace_func):
    ...     callable()

    '''
    trace_func_org = sys.gettrace()

    if thread:
        threading.settrace(trace_func)

    # Skip the context manager and this module otherwise they will be traced after
    # the yield is returned.
    trace_func = FilterByModuleName(trace_func, {contextmanager.__module__, __name__})
    sys.settrace(trace_func)

    try:
        yield
    finally:
        sys.settrace(trace_func_org)
        if thread:
            threading.settrace(trace_func_org)  # type: ignore



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from nextline.process.pdb.proxy import PdbInterfaceTraceFuncFactory

from .wrap import (
    AddFirstModule,
    DispatchForThreadOrTask,
    FilterByModuleName,
    FilterFirstModule,
    FilterLambda,
    FromFactory,
)

if TYPE_CHECKING:
    from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401

    from nextline.process.run import TraceContext


MODULES_TO_SKIP = {
    'multiprocessing.*',
    'threading',
    'queue',
    'importlib',
    'asyncio.*',
    'codec',
    'concurrent.futures.*',
    'selectors',
    'weakref',
    '_weakrefset',
    'socket',
    'logging',
    'os',
    'collections.*',
    'importlib.*',
    'pathlib',
    'typing',
    'posixpath',
    'fnmatch',
    '_pytest.*',
    'apluggy.*',
    'pluggy.*',
}


def Trace(context: TraceContext) -> TraceFunc:
    '''Build the main system trace function of Nextline.'''

    modules_to_trace = context['modules_to_trace']

    factory = FactoryForThreadOrTask(context=context)

    trace = DispatchForThreadOrTask(factory=factory)
    trace = AddFirstModule(modules_to_trace=modules_to_trace, trace=trace)
    trace = FilterLambda(trace=trace)
    trace = FilterByModuleName(patterns=MODULES_TO_SKIP, trace=trace)

    return trace


def FactoryForThreadOrTask(context: TraceContext) -> Callable[[], TraceFunc]:
    '''Return a function that returns a trace function for a thread or asyncio task.'''

    modules_to_trace = context['modules_to_trace']
    factory = PdbInterfaceTraceFuncFactory(context=context)

    def _trace_factory() -> TraceFunc:
        '''To be called in the thread or asyncio task to be traced.'''
        trace = FromFactory(factory=factory)
        trace = FilterFirstModule(trace=trace, modules_to_trace=modules_to_trace)
        return trace

    return _trace_factory



from __future__ import annotations

import threading
from asyncio import Task
from contextlib import contextmanager
from logging import getLogger
from queue import Queue
from threading import Thread
from typing import Dict, MutableMapping, Optional, Set
from weakref import WeakKeyDictionary

from apluggy import PluginManager

from nextline.count import PromptNoCounter, TraceNoCounter
from nextline.process.exc import TraceNotCalled
from nextline.process.io import peek_stdout_by_task_and_thread
from nextline.process.types import CommandQueueMap
from nextline.types import RunNo, TraceNo
from nextline.utils import (
    ThreadTaskDoneCallback,
    ThreadTaskIdComposer,
    current_task_or_thread,
)

from . import spec
from .plugins import (
    AddModuleToTrace,
    PromptInfoRegistrar,
    RegistrarProxy,
    StdoutRegistrar,
    TraceInfoRegistrar,
    TraceNumbersRegistrar,
    hookimpl,
)
from .types import TraceArgs


class TaskOrThreadToTraceMapper:
    def __init__(
        self, callback: 'Callback', trace_no_map: MutableMapping[Task | Thread, TraceNo]
    ) -> None:
        self._callback = callback
        self._trace_no_map = trace_no_map
        self._thread_task_done_callback = ThreadTaskDoneCallback(
            done=self._callback.task_or_thread_end
        )
        self._entering_thread: Optional[Thread] = None

        self._hook = PluginManager(spec.PROJECT_NAME)
        self._hook.add_hookspecs(spec)

    @hookimpl
    def task_or_thread_start(self, trace_no: TraceNo) -> None:
        task_or_thread = current_task_or_thread()
        self._trace_no_map[task_or_thread] = trace_no

        if task_or_thread is not self._entering_thread:
            self._thread_task_done_callback.register(task_or_thread)

        self._callback.trace_start(trace_no)

    @hookimpl
    def task_or_thread_end(self, task_or_thread: Task | Thread):
        trace_no = self._trace_no_map[task_or_thread]
        self._callback.trace_end(trace_no)

    @hookimpl
    def start(self) -> None:
        self._entering_thread = threading.current_thread()

    @hookimpl
    def close(self, exc_type=None, exc_value=None, traceback=None) -> None:
        self._thread_task_done_callback.close()
        if self._entering_thread:
            if trace_no := self._trace_no_map.get(self._entering_thread):
                self._callback.trace_end(trace_no)


class PeekStdout:
    def __init__(self, callback: 'Callback') -> None:
        self._callback = callback
        self._tasks_and_threads: Set[Task | Thread] = set()

    @hookimpl
    def task_or_thread_start(self) -> None:
        task_or_thread = current_task_or_thread()
        self._tasks_and_threads.add(task_or_thread)

    @hookimpl
    def start(self) -> None:
        self._peek_stdout = peek_stdout_by_task_and_thread(
            to_peek=self._tasks_and_threads, callback=self._callback.stdout
        )
        self._peek_stdout.__enter__()

    @hookimpl
    def close(self, exc_type=None, exc_value=None, traceback=None) -> None:
        self._peek_stdout.__exit__(exc_type, exc_value, traceback)


class Callback:
    def __init__(
        self,
        run_no: RunNo,
        registrar: RegistrarProxy,
        modules_to_trace: Set[str],
        command_queue_map: CommandQueueMap,
    ):
        self._trace_no_counter = TraceNoCounter(1)
        self._prompt_no_counter = PromptNoCounter(1)
        self._command_queue_map = command_queue_map
        self._trace_no_map: MutableMapping[Task | Thread, TraceNo] = WeakKeyDictionary()
        self._trace_args_map: Dict[TraceNo, TraceArgs] = {}
        self._trace_id_factory = ThreadTaskIdComposer()

        self._hook = PluginManager(spec.PROJECT_NAME)
        self._hook.add_hookspecs(spec)

        self._logger = getLogger(__name__)

        stdout_registrar = StdoutRegistrar(run_no=run_no, registrar=registrar)
        add_module_to_trace = AddModuleToTrace(modules_to_trace)
        trace_info_registrar = TraceInfoRegistrar(run_no=run_no, registrar=registrar)
        prompt_info_registrar = PromptInfoRegistrar(run_no=run_no, registrar=registrar)
        trace_numbers_registrar = TraceNumbersRegistrar(registrar=registrar)
        peek_stdout = PeekStdout(self)
        trace_mapper = TaskOrThreadToTraceMapper(self, self._trace_no_map)

        self._hook.register(stdout_registrar, name='stdout')
        self._hook.register(add_module_to_trace, name='add_module_to_trace')
        self._hook.register(trace_info_registrar, name='trace_info')
        self._hook.register(prompt_info_registrar, name='prompt_info')
        self._hook.register(trace_numbers_registrar, name='trace_numbers')
        self._hook.register(peek_stdout, name='peek_stdout')
        self._hook.register(trace_mapper, name='task_or_thread_to_trace_mapper')

    def task_or_thread_start(self) -> None:
        trace_no = self._trace_no_counter()
        self._hook.hook.task_or_thread_start(trace_no=trace_no)

    def task_or_thread_end(self, task_or_thread: Task | Thread):
        self._hook.hook.task_or_thread_end(task_or_thread=task_or_thread)

    def trace_start(self, trace_no: TraceNo):
        thread_task_id = self._trace_id_factory()
        thread_no = thread_task_id.thread_no
        task_no = thread_task_id.task_no
        self._command_queue_map[trace_no] = Queue()

        self._hook.hook.trace_start(
            trace_no=trace_no, thread_no=thread_no, task_no=task_no
        )

    def trace_end(self, trace_no: TraceNo):
        self._hook.hook.trace_end(trace_no=trace_no)
        del self._command_queue_map[trace_no]

    @contextmanager
    def trace_call(self, trace_args: TraceArgs):
        trace_no = self._trace_no_map[current_task_or_thread()]
        self._trace_args_map[trace_no] = trace_args
        with self._hook.with_.trace_call(trace_no=trace_no, trace_args=trace_args):
            try:
                yield
            finally:
                del self._trace_args_map[trace_no]

    @contextmanager
    def cmdloop(self):
        trace_no = self._trace_no_map[current_task_or_thread()]
        if (trace_args := self._trace_args_map.get(trace_no)) is None:
            raise TraceNotCalled
        with self._hook.with_.cmdloop(trace_no=trace_no, trace_args=trace_args):
            yield

    def prompt(self, text: str) -> str:
        trace_no = self._trace_no_map[current_task_or_thread()]
        prompt_no = self._prompt_no_counter()
        self._logger.debug(f'PromptNo: {prompt_no}')
        trace_args = self._trace_args_map[trace_no]
        with (
            p := self._hook.with_.prompt(
                trace_no=trace_no, prompt_no=prompt_no, trace_args=trace_args, out=text
            )
        ):
            while True:
                command, prompt_no_, trace_no_ = self._command_queue_map[trace_no].get()
                try:
                    assert trace_no_ == trace_no
                except AssertionError:
                    msg = f'TraceNo mismatch: {trace_no_} != {trace_no}'
                    self._logger.exception(msg)
                    raise
                if prompt_no_ == prompt_no:
                    break
                self._logger.warning(f'PromptNo mismatch: {prompt_no_} != {prompt_no}')
            p.gen.send(command)
        return command

    def stdout(self, task_or_thread: Task | Thread, line: str):
        trace_no = self._trace_no_map[task_or_thread]
        self._hook.hook.stdout(trace_no=trace_no, line=line)

    def start(self) -> None:
        self._hook.hook.start()

    def close(self, exc_type=None, exc_value=None, traceback=None) -> None:
        self._hook.hook.close(
            exc_type=exc_type, exc_value=exc_value, traceback=traceback
        )

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close(exc_type, exc_value, traceback)



from __future__ import annotations

from asyncio import Task
from collections.abc import MutableSet, Set
from functools import lru_cache, partial
from logging import getLogger
from threading import Thread
from types import FrameType
from typing import TYPE_CHECKING, Any, Callable, ContextManager, Iterable, Optional
from weakref import WeakKeyDictionary

from nextline.utils import current_task_or_thread, match_any

if TYPE_CHECKING:
    from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401


def FilterByModuleName(trace: TraceFunc, patterns: Iterable[str]) -> TraceFunc:
    '''Skip Python modules with names that match any of the patterns.'''

    patterns = frozenset(patterns)

    # NOTE: match_any() is slow
    match_any_ = lru_cache(partial(match_any, patterns=patterns))

    def filter(frame: FrameType, event, arg) -> bool:
        del event, arg
        module_name = frame.f_globals.get('__name__')
        return not match_any_(module_name)

    return Filter(trace=trace, filter=filter)


def FilterLambda(trace: TraceFunc) -> TraceFunc:
    '''Skip lambda functions.'''

    def filter(frame: FrameType, event, arg) -> bool:
        del event, arg
        func_name = frame.f_code.co_name
        return not func_name == '<lambda>'

    return Filter(trace=trace, filter=filter)


def Filter(
    trace: TraceFunc, filter: Callable[[FrameType, str, Any], bool]
) -> TraceFunc:
    '''Skip if the filter returns False.'''

    def _trace(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        if filter(frame, event, arg):
            return trace(frame, event, arg)
        return None

    return _trace


def AddFirstModule(trace: TraceFunc, modules_to_trace: MutableSet[str]) -> TraceFunc:
    '''Add the module name to the set the first time traced in a module with a name.'''

    def callback(frame: FrameType, event, arg) -> bool:
        del event, arg
        if module_name := frame.f_globals.get('__name__'):
            # The module has a name.
            modules_to_trace.add(module_name)
            logger = getLogger(__name__)
            msg = f'{AddFirstModule.__name__}: added {module_name!r}'
            logger.info(msg)
            return True
        return False

    return CallbackUntilAccepted(trace=trace, callback=callback)


def CallbackUntilAccepted(
    trace: TraceFunc, callback: Callable[[FrameType, str, object], bool]
) -> TraceFunc:
    '''Call the callback when traced until the callback returns True.'''
    first = True

    def create_local_trace() -> TraceFunc:
        '''Return a trace function that executes the callback.'''
        next_trace: TraceFunc | None = trace

        def local_trace(frame, event, arg) -> Optional[TraceFunc]:
            '''Execute the callback.'''
            nonlocal first, next_trace
            assert next_trace

            accepted = callback(frame, event, arg)

            logger = getLogger(__name__)
            name = CallbackUntilAccepted.__name__
            msg = f'{name}: the callback returned {accepted!r}'
            logger.debug(msg)

            if accepted:
                # Stop executing the callback.
                first = False
                return next_trace(frame, event, arg)

            # Continue until the callback accepts.
            if next_trace := next_trace(frame, event, arg):
                return local_trace
            return None

        return local_trace

    def global_trace(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        if not first:
            # The callback has already accepted.
            return trace(frame, event, arg)

        return create_local_trace()(frame, event, arg)

    return global_trace


def DispatchForThreadOrTask(factory: Callable[[], TraceFunc]) -> TraceFunc:
    '''Create a new trace function for each thread or asyncio task.'''

    map: WeakKeyDictionary[Task | Thread, TraceFunc] = WeakKeyDictionary()

    def ret(frame, event, arg) -> Optional[TraceFunc]:
        key = current_task_or_thread()
        trace = map.get(key)
        if not trace:
            trace = factory()
            map[key] = trace
            logger = getLogger(__name__)
            msg = f'{DispatchForThreadOrTask.__name__}: created trace for {key}'
            logger.info(msg)
        return trace(frame, event, arg)

    return ret


def FilterFirstModule(trace: TraceFunc, modules_to_trace: Set[str]) -> TraceFunc:
    '''Start tracing from a module in the set.

    Skip modules until reaching a module in the set. Stop skipping afterward.
    '''

    first = True

    def filter(frame: FrameType, event, arg) -> bool:
        nonlocal first
        del event, arg

        if first:
            module_name = frame.f_globals.get('__name__')
            if not match_any(module_name, modules_to_trace):
                return False

            first = False

            logger = getLogger(__name__)
            name = FilterFirstModule.__name__
            msg = f'{name}: started tracing at {module_name!r}'
            logger.info(msg)

        return True

    return Filter(trace=trace, filter=filter)


def FromFactory(factory: Callable[[], TraceFunc]) -> TraceFunc:
    '''Create a trace function first time called.

    Used to defer the creation of the trace function until need to call it.
    '''
    trace: TraceFunc | None = None

    def global_trace(frame, event, arg) -> Optional[TraceFunc]:
        nonlocal trace
        trace = trace or factory()
        return trace(frame, event, arg)

    return global_trace


def WithContext(
    trace: TraceFunc, context: Callable[[FrameType, str, Any], ContextManager[None]]
) -> TraceFunc:
    def _create_local_trace() -> TraceFunc:
        next_trace: TraceFunc | None = trace

        def _local_trace(frame, event, arg) -> Optional[TraceFunc]:
            nonlocal next_trace
            assert next_trace
            with context(frame, event, arg):
                if next_trace := next_trace(frame, event, arg):
                    return _local_trace
                return None

        return _local_trace

    def _global_trace(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        return _create_local_trace()(frame, event, arg)

    return _global_trace



__all__ = [
    'TaskDoneCallback',
    'ThreadDoneCallback',
    'ThreadTaskDoneCallback',
]

from .task import TaskDoneCallback
from .thread import ThreadDoneCallback
from .union import ThreadTaskDoneCallback



from __future__ import annotations

import asyncio
import functools
import threading
from typing import (
    TYPE_CHECKING,
    AsyncGenerator,
    AsyncIterator,
    Iterable,
    Set,
    Tuple,
    TypeVar,
)

if TYPE_CHECKING:
    from asyncio import Future, Task
    from threading import Thread


def current_task_or_thread() -> Task | Thread:
    """The asyncio task object if in a task, else the thread object"""
    try:
        task = asyncio.current_task()
    except RuntimeError:
        task = None
    return task or threading.current_thread()


try:
    from asyncio import to_thread  # type: ignore
except ImportError:
    # for Python 3.8
    # to_thread() is new in Python 3.9

    async def to_thread(func, /, *args, **kwargs):  # type: ignore
        loop = asyncio.get_running_loop()
        func_call = functools.partial(func, *args, **kwargs)
        return await loop.run_in_executor(None, func_call)


T = TypeVar("T")


async def agen_with_wait(
    agen: AsyncIterator[T],
) -> AsyncGenerator[
    T | Tuple[Tuple[Future, ...], Tuple[Future, ...]], Iterable[Future]
]:
    """Yield from the agen while waiting for received tasks

    Used to raise an exception from tasks
    """
    done: Set[Future] = set()
    pending: Set[Future] = set()
    anext = asyncio.ensure_future(agen.__anext__())
    while True:
        done_, pending_ = await asyncio.wait(
            pending | {anext},
            return_when=asyncio.FIRST_COMPLETED,
        )
        for t in done_ - {anext}:
            if exc := t.exception():
                anext.cancel()
                raise exc
        if anext in done_:
            try:
                item = anext.result()
            except StopAsyncIteration:
                break
            done_.remove(anext)
            done |= done_
        else:
            done |= done_
            continue
        pending &= pending_
        new = yield item
        if new is not None:
            pending |= set(new)
            yield tuple(done), tuple(pending)
            done.clear()
        anext = asyncio.ensure_future(agen.__anext__())


async def merge_aiters(
    *aiters: AsyncIterator[T],
) -> AsyncIterator[Tuple[int, T]]:
    aiter_map = {a: i for i, a in enumerate(aiters)}
    task_map = {asyncio.ensure_future(a.__anext__()): a for a in aiter_map.keys()}
    tasks = set(task_map.keys())
    while tasks:
        done, pending = await asyncio.wait(
            tasks,
            return_when=asyncio.FIRST_COMPLETED,
        )
        tasks = pending
        while done:
            task = done.pop()
            try:
                item = task.result()
            except StopAsyncIteration:
                continue
            aiter = task_map[task]
            yield aiter_map[aiter], item
            task = asyncio.ensure_future(aiter.__anext__())
            tasks.add(task)
            task_map[task] = aiter



from __future__ import annotations

import cProfile
import pstats
from io import StringIO
from typing import Callable, Tuple, TypeVar

_T = TypeVar("_T")


def profile_func(func: Callable[[], _T]) -> Tuple[str, _T]:
    with cProfile.Profile() as pr:
        ret = func()
    sortby = "cumulative"
    s = StringIO()
    pstats.Stats(pr, stream=s).strip_dirs().sort_stats(sortby).print_stats()
    return s.getvalue(), ret



class TraceNotCalled(RuntimeError):
    '''Pdb.cmdloop() was called without Callback.trace_call() being called first.

    This error can happen when sys.settrace() is called by Pdb, which will remove
    Nextline and Pdb is directly called by the system trace.

    This can happen, for example, when the user sends the Pdb command "step" at
    the very last line of the script.
    https://github.com/simonsobs/nextline/issues/1
    '''

    pass



from __future__ import annotations

from asyncio import Task
from threading import Thread
from typing import Any, Callable, Optional

from nextline.utils.func import current_task_or_thread, to_thread

from .task import TaskDoneCallback
from .thread import ThreadDoneCallback


class ThreadTaskDoneCallback:
    def __init__(
        self,
        done: Optional[Callable[[Task | Thread], Any]] = None,
        interval: float = 0.001,
    ):
        self._thread_callback = ThreadDoneCallback(done=done, interval=interval)
        self._task_callback = TaskDoneCallback(done=done)

    def register(self, task_or_thread: Optional[Task | Thread] = None) -> Task | Thread:
        if task_or_thread is None:
            task_or_thread = current_task_or_thread()
        if isinstance(task_or_thread, Task):
            self._task_callback.register(task_or_thread)
        else:
            self._thread_callback.register(task_or_thread)
        return task_or_thread

    def close(self, interval: float = 0.001) -> None:
        self._task_callback.close(interval=interval)
        self._thread_callback.close()

    async def aclose(self, interval: float = 0.001) -> None:
        """Awaitable version of close()"""
        await self._task_callback.aclose(interval=interval)
        await to_thread(self._thread_callback.close)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        self.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.aclose()



from __future__ import annotations

import fnmatch
from typing import Iterable


def match_any(filename: str | None, patterns: Iterable[str]) -> bool:
    '''Test if the filename matches any of the patterns

    This function is based on Bdb.is_skipped_module():
    https://github.com/python/cpython/blob/v3.9.5/Lib/bdb.py#L191
    '''
    if filename is None:
        return False
    return any(fnmatch.fnmatch(filename, pattern) for pattern in patterns)



from __future__ import annotations

import asyncio
import os
import signal
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures.process import BrokenProcessPool
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import partial
from logging import getLogger
from multiprocessing import Process
from typing import Callable, Generator, Generic, Tuple, TypeVar

from typing_extensions import TypeAlias

_T = TypeVar("_T")


@dataclass
class Result(Generic[_T]):
    '''An object returned by Running after the process has exited.'''

    returned: _T | None
    raised: BaseException | None
    process: Process
    process_created_at: datetime
    process_exited_at: datetime


class Running(Generic[_T]):
    '''An awaitable return value of `run_in_process()`.'''

    def __init__(
        self,
        process: Process,
        task: asyncio.Task[Tuple[_T | None, BaseException | None]],
    ):
        self.process = process
        self._task = task
        self.process_created_at = datetime.now(timezone.utc)
        self._process_created_at_fmt = self._format_time(self.process_created_at)
        self._log_created()

    def __repr__(self):
        ret = (
            f'<{self.__class__.__name__}'
            f' pid={self.process.pid!r}'
            f' created_at="{self._process_created_at_fmt}">'
        )
        return ret

    def _log_created(self) -> None:
        msg = f'Process ({self.process.pid}) created at {self._process_created_at_fmt}.'
        logger = getLogger()
        logger.info(msg)

    def _log_exited(self, exited_at: datetime) -> None:
        time_fmt = self._format_time(exited_at)
        exitcode = self.process.exitcode
        exit_fmt = f'{exitcode}'
        if exitcode and (name := _exitcode_to_name.get(exitcode)):
            exit_fmt = f'{exitcode} ({name})'
        pid = self.process.pid
        msg = f'Process ({pid}) exited at {time_fmt}. Exitcode: {exit_fmt}.'
        logger = getLogger(__name__)
        logger.info(msg)

    def _format_time(self, dt: datetime) -> str:
        return dt.strftime('%Y-%m-%d %H:%M:%S (%Z)')

    def interrupt(self) -> None:
        self.send_signal(signal.SIGINT)

    def send_signal(self, sig: int) -> None:
        if self.process.pid:
            os.kill(self.process.pid, sig)

    def terminate(self) -> None:
        self.process.terminate()

    def kill(self) -> None:
        self.process.kill()

    def __await__(self) -> Generator[None, None, Result[_T]]:
        # "yield from" in "__await__": https://stackoverflow.com/a/48261042/7309855
        ret, exc = yield from self._task.__await__()
        process_exited_at = datetime.now(timezone.utc)
        self._log_exited(process_exited_at)
        return Result(
            returned=ret,
            raised=exc,
            process=self.process,
            process_created_at=self.process_created_at,
            process_exited_at=process_exited_at,
        )


ExecutorFactory: TypeAlias = 'Callable[[], ProcessPoolExecutor]'


async def run_in_process(
    func: Callable[[], _T], executor_factory: ExecutorFactory | None = None
) -> Running[_T]:
    '''Call a function in a separate process and return an awaitable.

    Use functools.partial to pass arguments to the function.

    Example:

    >>> async def simple_example():
    ...
    ...     # Run pow(2, 3), which returns 8, in a separate process.
    ...     starting = run_in_process(partial(pow, 2, 3))
    ...
    ...     # Wait for the process to start.
    ...     running = await starting
    ...
    ...     # Wait for the process to finish.
    ...     result = await running
    ...
    ...     return result.returned

    >>> asyncio.run(simple_example())
    8

    '''

    if executor_factory is None:
        executor_factory = partial(ProcessPoolExecutor, max_workers=1)

    process: Process | None = None
    event = asyncio.Event()

    async def _run() -> Tuple[_T | None, BaseException | None]:
        nonlocal process
        assert executor_factory

        with executor_factory() as executor:
            loop = asyncio.get_running_loop()
            future = loop.run_in_executor(executor, func)
            process = list(executor._processes.values())[0]

            event.set()
            ret = None
            exc = None
            try:
                ret = await future
            except BrokenProcessPool:
                # NOTE: Not possible to use "as" for unknown reason.
                pass
            except BaseException as e:
                exc = e
            return ret, exc

    task = asyncio.create_task(_run())
    await event.wait()
    assert process
    ret = Running[_T](process=process, task=task)
    return ret


# Originally copied from
# https://github.com/python/cpython/blob/3.8/Lib/multiprocessing/process.py#L425-L429
_exitcode_to_name = {
    -signum: f'-{name}'
    for name, signum in signal.__dict__.items()
    if name[:3] == 'SIG' and '_' not in name
}



from __future__ import annotations

__all__ = [
    'QueueCommands',
    'QueueRegistry',
    'RunArg',
    'RunResult',
    'set_queues',
    'main',
]


from .run import run_
from .types import QueueCommands, QueueRegistry, RunArg, RunResult

_q_commands: QueueCommands | None = None
_q_registry: QueueRegistry | None = None


def set_queues(q_commands: QueueCommands, q_registry: QueueRegistry) -> None:
    '''Initializer of ProcessPoolExecutor that receives the queues.'''
    global _q_commands, _q_registry
    _q_commands = q_commands
    _q_registry = q_registry


def main(run_arg: RunArg) -> RunResult:
    '''The function to be submitted to ProcessPoolExecutor.'''
    assert _q_registry
    assert _q_commands
    return run_(run_arg, _q_commands, _q_registry)



from __future__ import annotations

from collections import defaultdict
from typing import AsyncIterator, DefaultDict, Generic, Optional, TypeVar

from .item import PubSubItem

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class PubSub(Generic[_KT, _VT]):
    """Asynchronous message broker of the publish-subscribe pattern"""

    def __init__(self) -> None:
        self._queue: DefaultDict[
            _KT,
            PubSubItem[_VT],
        ] = defaultdict(PubSubItem)

    def subscribe(self, key: _KT, last: Optional[bool] = True) -> AsyncIterator[_VT]:
        """Async iterator that yields values for the key as they are published

        Waits for new values and yields them as they are set. If `last` is
        true, yields immediately the latest value for the key before starting
        to wait. If the key doesn't exist, waits for the first value for the
        key; KeyError won't be raised.

        """
        return self._queue[key].subscribe(last=last)

    async def publish(self, key: _KT, value: _VT) -> None:
        """Yield the value in the generators"""
        await self._queue[key].publish(value)

    def latest(self, key: _KT) -> _VT:
        """Latest value for the key"""
        return self._queue[key].latest()

    async def end(self, key: _KT) -> None:
        """End all subscriptions for the key

        The async generators returned by the method `subscribe()` for the key
        will return.

        """
        if q := self._queue.pop(key, None):
            await q.close()

    async def close(self) -> None:
        """End all subscriptions for all keys

        All async generators returned by the method `subscribe()` for any key
        will return.

        """
        while self._queue:
            _, q = self._queue.popitem()
            await q.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.close()



from __future__ import annotations

import asyncio
import dataclasses
import datetime
from threading import Thread
from typing import MutableMapping  # noqa F401
from typing import TYPE_CHECKING, Optional

from typing_extensions import TypeAlias

from .types import TraceInfo  # noqa F401
from .types import RunInfo, RunNo
from .utils import PubSub, to_thread

if TYPE_CHECKING:
    from .process.run import QueueRegistry

SCRIPT_FILE_NAME = "<string>"

RunNoMap: TypeAlias = "MutableMapping[asyncio.Task | Thread, int]"
TraceNoMap: TypeAlias = "MutableMapping[asyncio.Task | Thread, int]"
TraceInfoMap: TypeAlias = "MutableMapping[int, TraceInfo]"


class Registrar:
    def __init__(self, registry: PubSub, queue: QueueRegistry):
        self._registry = registry
        self._queue = queue
        self._trace_info_map: TraceInfoMap = {}

    async def open(self):
        self._task = asyncio.create_task(self._relay())

    async def close(self):
        await to_thread(self._queue.put, None)
        await self._task

    async def _relay(self) -> None:
        while (m := await to_thread(self._queue.get)) is not None:
            key, value, close = m
            if close:
                await self._registry.end(key)
                continue
            if key == 'trace_info':
                assert isinstance(value, TraceInfo)
                self._trace_info_map[value.trace_no] = value
            await self._registry.publish(key, value)

    async def _end_traces(self) -> None:
        # In case the process is terminated or killed.
        for trace_no in self._registry.latest('trace_nos'):
            key = f'prompt_info_{trace_no}'
            if key in self._registry._queue:
                await self._registry.end(key)
            trace_info = self._trace_info_map[trace_no]
            if trace_info.state == "running":
                trace_info = dataclasses.replace(
                    trace_info,
                    state="finished",
                    ended_at=datetime.datetime.utcnow(),
                )
                await self._registry.publish('trace_info', trace_info)
        await self._registry.publish('trace_nos', tuple())
        # print({k: v._last_item for k, v in self._registry._queue.items()})

    async def script_change(self, script: str, filename: str) -> None:
        await self._registry.publish("statement", script)
        await self._registry.publish("script_file_name", filename)

    async def state_change(self, state_name: str) -> None:
        await self._registry.publish("state_name", state_name)

    async def state_initialized(self, run_no: int) -> None:
        await self._registry.publish("run_no", run_no)

    async def run_initialized(self, run_no: RunNo) -> None:
        self._run_info = RunInfo(
            run_no=run_no,
            state="initialized",
            script=self._registry.latest("statement"),
        )
        await self._registry.publish("run_info", self._run_info)

    async def run_start(self) -> None:
        self._run_info = dataclasses.replace(
            self._run_info,
            state="running",
            started_at=datetime.datetime.utcnow(),
        )
        await self._registry.publish("run_info", self._run_info)

    async def run_end(self, result: Optional[str], exception: Optional[str]) -> None:
        await self._end_traces()
        self._run_info = dataclasses.replace(
            self._run_info,
            state="finished",
            result=result,
            exception=exception,
            ended_at=datetime.datetime.utcnow(),
        )
        # TODO: check if run_no matches
        await self._registry.publish("run_info", self._run_info)



from __future__ import annotations

import enum
from asyncio import Condition, Queue
from typing import AsyncIterator, Generic, List, Literal, Optional, Tuple, TypeVar


class _M(enum.Enum):
    # TODO: Using Enum as sentinel for now as suggested in
    # https://stackoverflow.com/a/60605919/7309855. It still has a problem. For
    # example, the type of yielded values in subscribe() is not correctly
    # inferred as _T.

    START = object()
    END = object()


_T = TypeVar("_T")


class PubSubItem(Generic[_T]):
    '''Distribute items to asynchronous subscribers.

    Example:

    Define items to distribute:

    >>> items = ['2', '1', 'a', 'b', 'c']

    To demonstrate the "last" option of the method `subscribe()`, we distribute
    the first two items ("2" and "1") before starting subscribers. Subscribers
    with the "last" option "True" will immediately receive the most recent
    distributed item ("1") and then wait for new items. Subscribers with the
    "last" option "False" will wait for new items only.

    Define two subscribers, one with the "last" option "True" and the other with
    the "last" option "False":

    >>> async def subscriber_with_last(obj):
    ...     return [i async for i in obj.subscribe(last=True)]

    >>> async def subscriber_without_last(obj):
    ...     return [i async for i in obj.subscribe(last=False)]

    Define two distributors, one that distributes the first two items and the
    other that distributes the rest of the items:

    >>> async def distributor_first_two(obj):
    ...     for i in items[:2]:
    ...         await obj.publish(i)

    >>> async def distributor_rest(obj):
    ...     for i in items[2:]:
    ...         await obj.publish(i)
    ...     await obj.close()

    The second distributor calls the method `close()` to end the subscriptions.
    The method `subscribe()` will return when the method `close()` is called.

    The class can be instantiated without a running asyncio event loop:

    >>> obj = PubSubItem()

    Define the asynchronous main function:

    >>> async def main():
    ...     # Run the first distributor.
    ...     await distributor_first_two(obj)
    ...
    ...     # Run the two subscribers and the second distributor.
    ...     received1, received2, _ = await asyncio.gather(
    ...         subscriber_with_last(obj),
    ...         subscriber_without_last(obj),
    ...         distributor_rest(obj),
    ...     )
    ...
    ...     # Print the received items.
    ...     print(received1)
    ...     print(received2)

    Run the main function:

    >>> import asyncio
    >>> asyncio.run(main())
    ['1', 'a', 'b', 'c']
    ['a', 'b', 'c']

    '''

    def __init__(self) -> None:

        self._qs_out: List[Queue[Tuple[int, _T | Literal[_M.END]]]] = []
        self._last_enumerated: Tuple[int, _T | Literal[_M.START] | Literal[_M.END]] = (
            -1,
            _M.START,
        )

        self._last_item: _T | Literal[_M.START] = _M.START
        self._idx = -1

        self._closed: bool = False
        self._lock_close: Condition | None = None

    @property
    def nsubscriptions(self) -> int:
        """The number of the subscribers"""
        return len(self._qs_out)

    async def publish(self, item: _T) -> None:
        """Send data to subscribers"""
        if self._closed:
            raise RuntimeError(f"{self} is closed.")
        self._last_item = item
        await self._publish(item)

    def latest(self) -> _T:
        """Most recent data that have been published"""
        if self._last_item is _M.START:
            raise LookupError
        return self._last_item

    async def subscribe(self, last: Optional[bool] = True) -> AsyncIterator[_T]:
        """Yield data as they are put

        If `last` is true, yield immediately the most recent data before
        waiting for new data.
        """
        q: Queue[Tuple[int, _T | Literal[_M.END]]] = Queue()

        self._qs_out.append(q)

        try:
            last_idx, last_item = self._last_enumerated

            if last_item is _M.END:
                return

            if last and last_item is not _M.START:
                yield last_item

            while True:
                idx, item = await q.get()
                if item is _M.END:
                    break
                if last_idx < idx:
                    yield item

        finally:
            self._qs_out.remove(q)

    async def close(self) -> None:
        """End gracefully"""
        self._lock_close = self._lock_close or Condition()
        async with self._lock_close:
            if self._closed:
                return
            self._closed = True
            await self._publish(_M.END)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.close()

    async def _publish(self, item: _T | Literal[_M.END]) -> None:
        self._idx += 1
        self._last_enumerated = (self._idx, item)
        for q in list(self._qs_out):  # list in case it changes
            await q.put(self._last_enumerated)



from __future__ import annotations

from asyncio import Task, current_task
from collections import defaultdict
from threading import Thread, current_thread
from typing import Callable, DefaultDict, MutableMapping, Tuple
from weakref import WeakKeyDictionary

from nextline.count import TaskNoCounter, ThreadNoCounter
from nextline.types import TaskNo, ThreadNo, ThreadTaskId


class ThreadTaskIdComposer:
    """Create ThreadTaskId objects with unique thread and async task numbers"""

    def __init__(self) -> None:

        self.thread_no_counter = ThreadNoCounter(1)

        self._map: MutableMapping[Thread | Task, ThreadTaskId] = WeakKeyDictionary()

        self._thread_no_map: MutableMapping[Thread, ThreadNo] = WeakKeyDictionary()
        self._task_no_map: MutableMapping[Task, TaskNo] = WeakKeyDictionary()

        self._task_no_counter_map: DefaultDict[
            ThreadNo, Callable[[], TaskNo]
        ] = defaultdict(lambda: TaskNoCounter(1))

    def __call__(self) -> ThreadTaskId:
        """ThreadTaskId with the current thread and async task numbers

        Returns
        -------
        ThreadTaskId
            With the current thread and async task numbers. If not in an async
            task, the async task number will be None.
        """

        thread, task = self._current_thread_task()

        key = task or thread

        if ret := self._map.get(key):
            return ret

        ret = self._compose(thread, task)

        self._map[key] = ret

        return ret

    def has_id(self) -> bool:
        """True if the id has been created in the current thread and task"""
        thread, task = self._current_thread_task()
        key = task or thread
        return key in self._map

    def reset(self) -> None:
        """Start over the thread and async task numbers from one

        This method only resets the counters that generate thread and async
        task numbers. If the __call__() is executed in a thread and async task
        for which the ID has been generated before the reset, it will still
        return the same ID created before the reset.
        """
        self.thread_no_counter = ThreadNoCounter(1)
        self._task_no_counter_map.clear()

    def _current_thread_task(self) -> Tuple[Thread, Task | None]:
        try:
            task = current_task()
        except RuntimeError:
            task = None
        return current_thread(), task

    def _compose(self, thread: Thread, task: Task | None) -> ThreadTaskId:

        thread_no = self._thread_no_map.get(thread)
        if not thread_no:
            thread_no = self.thread_no_counter()
            assert thread_no
            self._thread_no_map[thread] = thread_no

        if not task:
            return ThreadTaskId(thread_no=thread_no, task_no=None)

        task_no = self._task_no_map.get(task)
        if not task_no:
            task_no = self._task_no_counter_map[thread_no]()
            self._task_no_map[task] = task_no

        return ThreadTaskId(thread_no=thread_no, task_no=task_no)



from __future__ import annotations

from io import TextIOWrapper
from logging import getLogger
from typing import Optional, Protocol


class _StdOut(TextIOWrapper):
    '''To be given to Pdb as stdout.'''

    def __init__(self, cli: CmdLoopInterface):
        self._cli = cli

    def write(self, s: str, /) -> int:
        '''Pdb prints user prompt or other messages to stdout.'''
        return self._cli.write(s)

    def flush(self):
        pass


class _StdIn(TextIOWrapper):
    '''To be given to Pdb as stdin.'''

    def __init__(self, cli: CmdLoopInterface):
        self._cli = cli

    def readline(self):
        '''Pdb waits for user input from stdin.'''
        return self._cli.prompt()


class PromptFunc(Protocol):
    def __call__(self, text: str) -> str:
        ...


class CmdLoopInterface:
    def __init__(self, prompt_func: PromptFunc) -> None:

        # Must be assigned to Pdb.prompt.
        self.prompt_end: Optional[str] = None

        # To be given to Pdb as stdout and stdin.
        self.stdout = _StdOut(self)
        self.stdin = _StdIn(self)

        self._prompt = prompt_func
        self._prompt_text = ''

    def write(self, s: str) -> int:
        '''Called by _StdOut.write()'''
        self._prompt_text += s
        return len(s)

    def prompt(self) -> str:
        '''Called by _StdIn.readline()'''

        prompt_text = self._prompt_text

        logger = getLogger(__name__)

        logger.debug(f'Prompt text: {prompt_text!r}')

        try:
            assert self.prompt_end is not None
        except AssertionError:
            logger.exception(f'{self.__class__.__name__!r} has no prompt_end')
            raise

        if not prompt_text.endswith(self.prompt_end):
            msg = f'{prompt_text!r} does not end with {self.prompt_end!r}'
            logger.warning(msg)

        self._prompt_text = ''

        command = self._prompt(text=prompt_text)

        return command



"""The creation of a callable from code to be traced.

The name of this module becomes the module name of the callable.
"""

from __future__ import annotations

from functools import partial
from types import CodeType
from typing import Callable


def compose(code: CodeType | str) -> Callable[[], None]:
    """Create a function that executes the code

    Parameters
    ----------
    code : object
        A code to be executed when the returned function is called.
        This must be an object that can be executed by the Python
        built-in Function exec().

    Returns
    -------
    function
        A function with no arguments that executes the code

    """
    globals_ = {"__name__": __name__}
    # The globals, the second arg of exec(), resolves the issue*. An
    # empty dict would suffice for the issus.
    #
    # The __name__ is given for a different purpose. It becomes the
    # module name of the returned function. The module name is used in
    # Trace.
    #
    # [*] https://github.com/simonsobs/nextline/issues/7

    func = partial(exec, code, globals_)
    return func



from __future__ import annotations

import dataclasses
import datetime
from typing import NewType, Optional

RunNo = NewType("RunNo", int)
TraceNo = NewType("TraceNo", int)
ThreadNo = NewType("ThreadNo", int)
TaskNo = NewType("TaskNo", int)
PromptNo = NewType("PromptNo", int)


@dataclasses.dataclass(frozen=True)
class ThreadTaskId:
    thread_no: ThreadNo
    task_no: TaskNo | None


@dataclasses.dataclass(frozen=True)
class RunInfo:
    run_no: RunNo
    state: str
    script: Optional[str] = None
    result: Optional[str] = None
    exception: Optional[str] = None
    started_at: Optional[datetime.datetime] = None
    ended_at: Optional[datetime.datetime] = None


@dataclasses.dataclass(frozen=True)
class TraceInfo:
    run_no: RunNo
    state: str
    trace_no: TraceNo
    thread_no: ThreadNo
    task_no: Optional[TaskNo] = None
    started_at: Optional[datetime.datetime] = None
    ended_at: Optional[datetime.datetime] = None


@dataclasses.dataclass(frozen=True)
class PromptInfo:
    run_no: RunNo
    trace_no: TraceNo
    prompt_no: PromptNo
    open: bool
    event: Optional[str] = None
    file_name: Optional[str] = None
    line_no: Optional[int] = None
    stdout: Optional[str] = None
    command: Optional[str] = None
    started_at: Optional[datetime.datetime] = None
    ended_at: Optional[datetime.datetime] = None
    trace_call_end: Optional[bool] = False  # TODO: remove when possible


@dataclasses.dataclass(frozen=True)
class StdoutInfo:
    run_no: RunNo
    trace_no: TraceNo
    text: Optional[str] = None
    written_at: Optional[datetime.datetime] = None



import atexit
import sys
from concurrent.futures.process import _python_exit
from typing import Any, Callable, List


def suppress_atexit_oserror():
    '''Catch the OS error at exit in Python 3.8

    The following error can occur at the end of the pytest in Python 3.8:

      Error in atexit._run_exitfuncs:
      Traceback (most recent call last):
        File ...
          ...
        File "... lib/python3.8/multiprocessing/connection.py", line 136, in _check_closed
          raise OSError("handle is closed")
      OSError: handle is closed

    This function is used as a fixture to suppress the error.

    Usage:
      Add the following code to the conftest.py file:

        import pytest
        from nextline.test import suppress_atexit_oserror
        _ = pytest.fixture(scope='session', autouse=True)(suppress_atexit_oserror)


    Reference: https://stackoverflow.com/q/68091084/7309855

    '''
    yield
    if sys.version_info >= (3, 9):
        return

    if not _atexit_is_registered(_python_exit):
        return

    atexit.unregister(_python_exit)
    atexit.register(_wrap_python_exit)


def _wrap_python_exit(*args, **kwargs):
    try:
        _python_exit(*args, **kwargs)
    except OSError:
        pass


def _atexit_is_registered(func: Callable[..., Any]) -> bool:
    return func in _atexit_registered()


def _atexit_registered() -> List[Callable[..., Any]]:
    # https://stackoverflow.com/a/63813607/7309855
    import atexit

    ret = []

    class Capture:
        def __eq__(self, other):
            ret.append(other)
            return False

        def __call__(self):
            pass

    c = Capture()
    atexit.unregister(c)
    return ret



from __future__ import annotations

from itertools import count
from typing import Callable, Type, TypeVar

from nextline.types import PromptNo, RunNo, TaskNo, ThreadNo, TraceNo

_T = TypeVar("_T", bound=int)


def RunNoCounter(start=1) -> Callable[[], RunNo]:
    return CastedCounter(count(start).__next__, RunNo)


def TraceNoCounter(start=1) -> Callable[[], TraceNo]:
    return CastedCounter(count(start).__next__, TraceNo)


def ThreadNoCounter(start=1) -> Callable[[], ThreadNo]:
    return CastedCounter(count(start).__next__, ThreadNo)


def TaskNoCounter(start=1) -> Callable[[], TaskNo]:
    return CastedCounter(count(start).__next__, TaskNo)


def PromptNoCounter(start=1) -> Callable[[], PromptNo]:
    return CastedCounter(count(start).__next__, PromptNo)


def CastedCounter(src: Callable[[], int], type_: Type[_T]) -> Callable[[], _T]:
    def casted_counter() -> _T:
        return type_(src())

    return casted_counter



from __future__ import annotations

from logging import getLogger
from pdb import Pdb
from typing import IO, Callable, ContextManager

from nextline.process.exc import TraceNotCalled


class CustomizedPdb(Pdb):
    '''A Pdb subclass that hooks the command loop.'''

    def __init__(
        self,
        cmdloop_hook: Callable[[], ContextManager[None]],
        stdin: IO[str],
        stdout: IO[str],
    ):
        super().__init__(stdin=stdin, stdout=stdout, nosigint=True, readrc=False)
        # NOTE: nosigint (No SIGINT) is False by default.  When False, Pdb lets
        # the user to break into a debugger again with SIGINT after the
        # `continue` command is issued. This feature is potentially useful in
        # Nextline.
        #
        #  However, it is turned off now for two reasons:
        #
        #  1. Pdb calls `sys.settrace(None)` when the `continue` command is
        #     issued and `sys.settrace(pdb.trace_dispatch)` when SIGINT is
        #     handled, which removes Nextline.  We will need to override Pdb
        #     methods to set Nextline back.
        #
        #  2. Nextline uses multiple instances of Pdb.  We will need to break
        #     into the right instance of Pdb when SIGINT is handled.

        self._cmdloop_hook = cmdloop_hook

        # self.quitting = True # not sure if necessary

        # stop at the first line
        self.botframe = None
        self._set_stopinfo(None, None)  # type: ignore

    def _cmdloop(self) -> None:
        '''Override Pdb._cmdloop() to keep it from catching KeyboardInterrupt.'''
        # super()._cmdloop()
        self.cmdloop()

    def cmdloop(self, intro=None) -> None:
        '''Override Cmd.cmdloop() to call it inside a context manager.'''
        try:
            with self._cmdloop_hook():
                super().cmdloop(intro=intro)
        except TraceNotCalled:
            logger = getLogger(__name__)
            logger.exception('')

    def set_continue(self):
        '''Override Bdb.set_continue() to avoid sys.settrace(None).'''
        # super().set_continue()
        self._set_stopinfo(self.botframe, None, -1)



from __future__ import annotations

import time
from threading import Thread, current_thread
from typing import Any, Callable, Optional, Set

from nextline.utils.thread_exception import ExcThread


class ThreadDoneCallback:
    """Call a function when each registered thread ends

    Parameters
    ----------
    done : callable, optional
        A function with one arg. Each time a registered thread ends, this
        function will be called with the thread object as the arg. The return
        value will be ignored.

        The `done` is optional. This class can be still useful to wait for all
        registered tasks to end.

    interval : float, default 0.001 [sec]
        The period of determining if any registered thread ends.
    """

    def __init__(
        self,
        done: Optional[Callable[[Thread], Any]] = None,
        interval: float = 0.001,
    ):
        self._done = done
        self._interval = interval

        self._active: Set[Thread] = set()
        self._closed = False

        self._t = ExcThread(target=self._monitor, daemon=True)
        self._t.start()

    def register(self, thread: Optional[Thread] = None) -> Thread:
        """Add the current thread by default, or the given thread

        The callback function `done`, given at the initialization,
        will be called with the thread object when the thread ends.
        """
        if thread is None:
            thread = current_thread()
        self._active.add(thread)
        return thread

    def close(self) -> None:
        """To be optionally called after all threads are registered

        This method returns after all registered threads end.

        The method cannot be called from a registered thread.

        If exceptions are raised in the callback function, this method
        re-raises the first exception.

        """
        if current_thread() in self._active:
            raise RuntimeError("The close() cannot be called from a registered thread")
        self._closed = True
        self._t.join()

    def _monitor(self):
        exc = []
        while True:
            if done := {t for t in self._active if not t.is_alive()}:
                if self._done:
                    for d in done:
                        try:
                            self._done(d)
                        except BaseException as e:
                            exc.append(e)
                self._active = self._active - done
            time.sleep(self._interval)
            if self._active:
                continue
            if self._closed:
                break
        if exc:
            raise exc[0]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        self.close()



from threading import Thread


class ExcThread(Thread):
    """Wrap Thread so to raise exception at join

    The implementation based on
    https://www.geeksforgeeks.org/handling-a-threads-exception-in-the-caller-thread-in-python/

    """

    def run(self):
        self.exc = None
        try:
            return super().run()
        except BaseException as e:
            self.exc = e

    def join(self):
        ret = super().join()
        if self.exc:
            raise self.exc
        return ret



from __future__ import annotations

from asyncio import Task
from threading import Thread
from typing import Generator, Optional

import apluggy as pluggy
from apluggy import contextmanager

from nextline.types import PromptNo, TaskNo, ThreadNo, TraceNo

from .types import TraceArgs

PROJECT_NAME = 'nextline_process_callback'


hookspec = pluggy.HookspecMarker(PROJECT_NAME)
hookimpl = pluggy.HookimplMarker(PROJECT_NAME)


@hookspec
def task_or_thread_start(trace_no: TraceNo) -> None:
    pass


@hookspec
def task_or_thread_end(task_or_thread: Task | Thread):
    pass


@hookspec
def trace_start(
    trace_no: TraceNo, thread_no: ThreadNo, task_no: Optional[TaskNo]
) -> None:
    pass


@hookspec
def trace_end(trace_no: TraceNo) -> None:
    pass


@hookspec
@contextmanager
def trace_call(trace_no: TraceNo, trace_args: TraceArgs):
    pass


@hookspec
@contextmanager
def cmdloop(trace_no: TraceNo, trace_args: TraceArgs):
    pass


@hookspec
@contextmanager
def prompt(
    trace_no: TraceNo, prompt_no: PromptNo, trace_args: TraceArgs, out: str
) -> Generator[None, str, None]:
    # Receive the command by gen.send().
    command = yield  # noqa: F841
    yield


@hookspec
def stdout(trace_no: TraceNo, line: str) -> None:
    pass


@hookspec
def start() -> None:
    pass


@hookspec
def close(exc_type=None, exc_value=None, traceback=None) -> None:
    pass



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from logging import getLogger
from types import CodeType
from typing import Any, Callable, Set, TypedDict, TypeVar

from nextline.types import RunNo

from . import script
from .call import sys_trace
from .callback import Callback, RegistrarProxy
from .trace import Trace
from .types import CommandQueueMap, QueueCommands, QueueRegistry, RunArg, RunResult

_T = TypeVar('_T')


class TraceContext(TypedDict):
    callback: Callback
    modules_to_trace: Set[str]


def run_(
    run_arg: RunArg, q_commands: QueueCommands, q_registry: QueueRegistry
) -> RunResult:

    run_no = run_arg['run_no']

    statement = run_arg['statement']
    filename = run_arg['filename']

    try:
        code = _compile(statement, filename)
    except BaseException as e:
        return RunResult(ret=None, exc=e)

    func = script.compose(code)

    return run_with_trace(run_no, func, q_commands, q_registry)


def run_with_trace(
    run_no: RunNo,
    func: Callable[[], Any],
    q_commands: QueueCommands,
    q_registry: QueueRegistry,
) -> RunResult:

    ret: Any = None
    exc: BaseException | None = None

    with _trace(run_no, q_commands, q_registry) as trace:
        with sys_trace(trace_func=trace):
            try:
                ret = func()
            except BaseException as e:
                exc = e

    # NOTE: How to print the exception in the same way as the interpreter.
    # import traceback
    # traceback.print_exception(type(exc), exc, exc.__traceback__)

    if exc and exc.__traceback__:
        # remove this frame from the traceback.
        # Note: exc.__traceback__ is sys._getframe()
        exc.__traceback__ = exc.__traceback__.tb_next

    return RunResult(ret=ret, exc=exc)


@contextmanager
def _trace(run_no: RunNo, q_commands: QueueCommands, q_registry: QueueRegistry):

    command_queue_map: CommandQueueMap = {}
    modules_to_trace: Set[str] = set()

    with Callback(
        run_no=run_no,
        registrar=RegistrarProxy(q_registry),
        modules_to_trace=modules_to_trace,
        command_queue_map=command_queue_map,
    ) as callback:

        context = TraceContext(
            callback=callback,
            modules_to_trace=modules_to_trace,
        )

        trace = Trace(context=context)

        with relay_commands(q_commands, command_queue_map):
            yield trace


def _compile(code: CodeType | str, filename: str) -> CodeType:
    if isinstance(code, str):
        return compile(code, filename, 'exec')
    return code


@contextmanager
def relay_commands(q_commands: QueueCommands, command_queue_map: CommandQueueMap):
    '''Pass the Pdb commands from the main process to the Pdb instances.'''
    logger = getLogger(__name__)

    def fn() -> None:
        assert q_commands
        while m := q_commands.get():
            logger.debug(f'q_commands.get() -> {m!r}')
            command, prompt_no, trace_no = m
            command_queue_map[trace_no].put(m)

    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(try_again_on_error, fn)  # type: ignore
        try:
            yield
        finally:
            q_commands.put(None)
            future.result()


def try_again_on_error(func: Callable[[], _T]) -> _T:
    '''Keep trying until the function succeeds without an exception.'''
    while True:
        try:
            return func()
        # except KeyboardInterrupt:
        #     raise
        except BaseException:
            logger = getLogger(__name__)
            logger.exception('')



from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, Callable

from nextline.process.trace.wrap import WithContext

from .custom import CustomizedPdb
from .stream import CmdLoopInterface

if TYPE_CHECKING:
    from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401

    from nextline.process.run import TraceContext


def PdbInterfaceTraceFuncFactory(context: TraceContext) -> Callable[[], TraceFunc]:
    def factory() -> TraceFunc:

        callback = context['callback']
        callback.task_or_thread_start()
        trace = instantiate_pdb(context=context)

        trace = TraceCallCallback(trace=trace, context=context)
        # TODO: Add a test. The tests pass without the above line.  Without it,
        #       the arrow in the web UI does not move when the Pdb is "continuing."

        return trace

    return factory


def TraceCallCallback(trace: TraceFunc, context: TraceContext) -> TraceFunc:
    callback = context['callback']

    @contextmanager
    def _context(frame, event, arg):
        with callback.trace_call(trace_args=(frame, event, arg)):
            yield

    return WithContext(trace, context=_context)


def instantiate_pdb(context: TraceContext):
    '''Create a new Pdb instance with callback hooked and return its trace function.'''

    cmdloop_interface = CmdLoopInterface(prompt_func=context['callback'].prompt)

    pdb = CustomizedPdb(
        cmdloop_hook=context['callback'].cmdloop,
        stdin=cmdloop_interface.stdin,
        stdout=cmdloop_interface.stdout,
    )
    cmdloop_interface.prompt_end = pdb.prompt

    return pdb.trace_dispatch



from __future__ import annotations

from asyncio import Task  # noqa F401
from collections import defaultdict
from contextlib import contextmanager
from threading import Thread  # noqa F401
from typing import Any, Callable, Collection, DefaultDict, TypeVar

from typing_extensions import TypeAlias

from nextline.utils import current_task_or_thread, peek_stdout

_T = TypeVar("_T")
_Key: TypeAlias = "Task | Thread"


@contextmanager
def peek_stdout_by_task_and_thread(
    to_peek: Collection[_Key],
    callback: Callable[[_Key, str], Any],
):
    key_factory = CurrentTaskOrThreadIfInCollection(collection=to_peek)
    callback_ = ReadLinesByKey(callback)
    assign_key = AssignKey(key_factory=key_factory, callback=callback_)  # type: ignore
    with peek_stdout(assign_key) as t:
        yield t


def CurrentTaskOrThreadIfInCollection(
    collection: Collection[_Key],
) -> Callable[[], _Key | None]:
    def fn() -> _Key | None:
        if (key := current_task_or_thread()) in collection:
            return key
        return None

    return fn


def ReadLinesByKey(callback: Callable[[_T, str], Any]) -> Callable[[_T, str], Any]:
    buffer: DefaultDict[_T, str] = defaultdict(str)

    def read_lines_by_key(key: _T, s: str) -> None:
        buffer[key] += s
        if s.endswith("\n"):
            line = buffer.pop(key)
            callback(key, line)

    return read_lines_by_key


def AssignKey(
    key_factory: Callable[[], _T | None], callback: Callable[[_T, str], Any]
) -> Callable[[str], None]:
    def assign_key(s: str) -> None:
        if key := key_factory():
            callback(key, s)

    return assign_key



from typing import Any, Optional

from transitions import EventData

from nextline.context import Context

from .factory import build_state_machine


class Machine:
    '''The finite state machine of the nextline states.'''

    def __init__(self, context: Context):
        self._context = context

        self._machine = build_state_machine(model=self)
        self._machine.after_state_change = self.after_state_change.__name__  # type: ignore

        assert self.state  # type: ignore

    def __repr__(self):
        # e.g., "<Machine 'running'>"
        return f'<{self.__class__.__name__} {self.state!r}>'

    async def after_state_change(self, _: EventData) -> None:
        await self._context.state_change(self.state)  # type: ignore

    async def on_enter_initialized(self, _: EventData) -> None:
        await self._context.initialize()

    async def on_enter_running(self, _: EventData) -> None:
        await self._context.run()

    async def on_exit_running(self, _: EventData) -> None:
        await self._context.finish()
        # NOTE: Call context.finish() here rather than in on_enter_finished()
        # because the state should be "running" until context.finish() returns.
        # The on_enter_finished() callback is called after the state is changed.

    def send_pdb_command(self, command: str, prompt_no: int, trace_no: int) -> None:
        assert self.is_running()  # type: ignore
        self._context.send_pdb_command(command, prompt_no, trace_no)

    def interrupt(self) -> None:
        assert self.is_running()  # type: ignore
        self._context.interrupt()

    def terminate(self) -> None:
        assert self.is_running()  # type: ignore
        self._context.terminate()

    def kill(self) -> None:
        assert self.is_running()  # type: ignore
        self._context.kill()

    def exception(self) -> Optional[BaseException]:
        assert self.is_finished()  # type: ignore
        return self._context.exception()

    def result(self) -> Any:
        assert self.is_finished()  # type: ignore
        return self._context.result()

    async def on_enter_closed(self, _: EventData) -> None:
        await self._context.close()

    async def on_reset(self, event: EventData) -> None:
        await self._context.reset(*event.args, **event.kwargs)

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.close()



from __future__ import annotations

import dataclasses
import datetime
import os
from contextlib import _GeneratorContextManager
from logging import getLogger
from types import FrameType
from typing import TYPE_CHECKING, Callable, Dict, Generator, Optional, Set, Tuple

from apluggy import contextmanager

from nextline.types import (
    PromptInfo,
    PromptNo,
    RunNo,
    StdoutInfo,
    TaskNo,
    ThreadNo,
    TraceInfo,
    TraceNo,
)

from .spec import hookimpl
from .types import TraceArgs

if TYPE_CHECKING:
    from ..run import QueueRegistry


class TraceNumbersRegistrar:
    def __init__(self, registrar: RegistrarProxy):
        self._registrar = registrar
        self._trace_nos: Tuple[TraceNo, ...] = ()

    @hookimpl
    def trace_start(self, trace_no: TraceNo):
        self._trace_nos = self._trace_nos + (trace_no,)
        self._registrar.put_trace_nos(self._trace_nos)

    @hookimpl
    def trace_end(self, trace_no: TraceNo):
        nosl = list(self._trace_nos)
        nosl.remove(trace_no)
        self._trace_nos = tuple(nosl)
        self._registrar.put_trace_nos(self._trace_nos)


class TraceInfoRegistrar:
    def __init__(self, run_no: RunNo, registrar: RegistrarProxy):
        self._run_no = run_no
        self._registrar = registrar
        self._trace_context_map: Dict[TraceNo, _GeneratorContextManager] = {}

    @hookimpl
    def trace_start(
        self,
        trace_no: TraceNo,
        thread_no: ThreadNo,
        task_no: Optional[TaskNo],
    ) -> None:
        @contextmanager
        def _trace():

            trace_info = TraceInfo(
                run_no=self._run_no,
                trace_no=trace_no,
                thread_no=thread_no,
                task_no=task_no,
                state="running",
                started_at=datetime.datetime.utcnow(),
            )
            self._registrar.put_trace_info(trace_info)

            yield

            trace_info_end = dataclasses.replace(
                trace_info,
                state='finished',
                ended_at=datetime.datetime.utcnow(),
            )
            self._registrar.put_trace_info(trace_info_end)

        context = _trace()
        context.__enter__()
        self._trace_context_map[trace_no] = context

    @hookimpl
    def trace_end(self, trace_no: TraceNo) -> None:
        context = self._trace_context_map.pop(trace_no)
        context.__exit__(None, None, None)


class PromptInfoRegistrar:
    def __init__(self, run_no: RunNo, registrar: RegistrarProxy):
        self._run_no = run_no
        self._registrar = registrar
        self._trace_context_map: Dict[TraceNo, _GeneratorContextManager] = {}
        # self._trace_call_context_map: Dict[TraceNo, _GeneratorContextManager] = {}
        # self._prompt_context_map: Dict[PromptNo, _GeneratorContextManager] = {}
        self._last_prompt_frame_map: Dict[TraceNo, FrameType] = {}

    @hookimpl
    def trace_start(self, trace_no: TraceNo) -> None:
        @contextmanager
        def _trace():

            # TODO: Putting a prompt info for now because otherwise tests get stuck
            # sometimes for an unknown reason. Need to investigate
            prompt_info = PromptInfo(
                run_no=self._run_no,
                trace_no=trace_no,
                prompt_no=PromptNo(-1),
                open=False,
            )
            self._registrar.put_prompt_info_for_trace(trace_no, prompt_info)

            yield

            self._registrar.end_prompt_info_for_trace(trace_no)

        context = _trace()
        context.__enter__()
        self._trace_context_map[trace_no] = context

    @hookimpl
    def trace_end(self, trace_no: TraceNo) -> None:
        context = self._trace_context_map.pop(trace_no)
        context.__exit__(None, None, None)

    @hookimpl
    @contextmanager
    def trace_call(self, trace_no: TraceNo, trace_args: TraceArgs):

        try:
            yield
        finally:
            frame, event, _ = trace_args
            file_name = _to_canonic(frame.f_code.co_filename)
            line_no = frame.f_lineno

            if frame is not self._last_prompt_frame_map.get(trace_no):
                return

            # TODO: Sending a prompt info with "open=False" for now so that the
            #       arrow in the web UI moves when the Pdb is "continuing."

            prompt_info = PromptInfo(
                run_no=self._run_no,
                trace_no=trace_no,
                prompt_no=PromptNo(-1),
                open=False,
                event=event,
                file_name=file_name,
                line_no=line_no,
                trace_call_end=True,
            )
            self._registrar.put_prompt_info(prompt_info)
            self._registrar.put_prompt_info_for_trace(trace_no, prompt_info)

    @hookimpl
    @contextmanager
    def prompt(
        self,
        trace_no: TraceNo,
        prompt_no: PromptNo,
        trace_args: TraceArgs,
        out: str,
    ) -> Generator[None, str, None]:

        frame, event, _ = trace_args
        file_name = _to_canonic(frame.f_code.co_filename)
        line_no = frame.f_lineno
        prompt_info = PromptInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            prompt_no=prompt_no,
            open=True,
            event=event,
            file_name=file_name,
            line_no=line_no,
            stdout=out,
            started_at=datetime.datetime.utcnow(),
        )
        self._registrar.put_prompt_info(prompt_info)
        self._registrar.put_prompt_info_for_trace(trace_no, prompt_info)

        self._last_prompt_frame_map[trace_no] = frame

        # Yield twice: once to receive from send(), and once to exit.
        # https://stackoverflow.com/a/68304565/7309855
        command = yield
        yield

        prompt_info_end = dataclasses.replace(
            prompt_info,
            open=False,
            command=command,
            ended_at=datetime.datetime.utcnow(),
        )
        self._registrar.put_prompt_info(prompt_info_end)
        self._registrar.put_prompt_info_for_trace(trace_no, prompt_info_end)


class AddModuleToTrace:
    '''Let Python modules be traced in new threads and asyncio tasks.'''

    def __init__(self, modules_to_trace: Set[str]):
        self._modules_to_trace = modules_to_trace
        self._logger = getLogger(__name__)

    @hookimpl
    @contextmanager
    def cmdloop(self, trace_args: TraceArgs) -> Generator[None, str, None]:
        frame, _, _ = trace_args
        if module_name := frame.f_globals.get('__name__'):
            msg = f'{self.__class__.__name__}: adding {module_name!r}'
            self._logger.info(msg)
            self._modules_to_trace.add(module_name)
        yield


class StdoutRegistrar:
    def __init__(self, run_no: RunNo, registrar: RegistrarProxy):
        self._run_no = run_no
        self._registrar = registrar

    @hookimpl
    def stdout(self, trace_no: TraceNo, line: str):
        stdout_info = StdoutInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            text=line,
            written_at=datetime.datetime.utcnow(),
        )
        self._registrar.put_stdout_info(stdout_info)


class RegistrarProxy:
    def __init__(self, queue: QueueRegistry):
        self._queue = queue

    def put_trace_nos(self, trace_nos: Tuple[TraceNo, ...]) -> None:
        self._queue.put(("trace_nos", trace_nos, False))

    def put_trace_info(self, trace_info: TraceInfo) -> None:
        self._queue.put(("trace_info", trace_info, False))

    def put_prompt_info(self, prompt_info: PromptInfo) -> None:
        self._queue.put(("prompt_info", prompt_info, False))

    def put_prompt_info_for_trace(
        self, trace_no: TraceNo, prompt_info: PromptInfo
    ) -> None:
        key = f"prompt_info_{trace_no}"
        self._queue.put((key, prompt_info, False))

    def end_prompt_info_for_trace(self, trace_no: TraceNo) -> None:
        key = f"prompt_info_{trace_no}"
        self._queue.put((key, None, True))

    def put_stdout_info(self, stdout_info: StdoutInfo) -> None:
        self._queue.put(("stdout", stdout_info, False))


def ToCanonic() -> Callable[[str], str]:
    # Based on Bdb.canonic()
    # https://github.com/python/cpython/blob/v3.10.5/Lib/bdb.py#L39-L54

    cache: Dict[str, str] = {}

    def to_canonic(filename: str) -> str:
        if filename == "<" + filename[1:-1] + ">":
            return filename
        canonic = cache.get(filename)
        if not canonic:
            canonic = os.path.abspath(filename)
            canonic = os.path.normcase(canonic)
            cache[filename] = canonic
        return canonic

    return to_canonic


_to_canonic = ToCanonic()



import json
import traceback
from dataclasses import dataclass, field
from queue import Queue
from typing import Any, MutableMapping, Optional, Tuple, TypedDict

from typing_extensions import TypeAlias

from nextline.types import PromptNo, RunNo, TraceNo

PdbCommand: TypeAlias = str
QueueCommands: TypeAlias = "Queue[Tuple[PdbCommand, PromptNo, TraceNo] | None]"
QueueRegistry: TypeAlias = "Queue[Tuple[str, Any, bool]]"

CommandQueueMap: TypeAlias = MutableMapping[
    TraceNo, 'Queue[Tuple[PdbCommand, PromptNo, TraceNo]]'
]


class RunArg(TypedDict):
    run_no: RunNo
    statement: str
    filename: str


@dataclass
class RunResult:
    ret: Optional[Any]
    exc: Optional[BaseException]
    _fmt_ret: Optional[str] = field(init=False, repr=False, default=None)
    _fmt_exc: Optional[str] = field(init=False, repr=False, default=None)

    @property
    def fmt_ret(self) -> str:
        if self._fmt_ret is None:
            self._fmt_ret = json.dumps(self.ret)
        return self._fmt_ret

    @property
    def fmt_exc(self) -> str:
        if self._fmt_exc is None:
            if self.exc is None:
                self._fmt_exc = ''
            else:
                self._fmt_exc = ''.join(
                    traceback.format_exception(
                        type(self.exc),
                        self.exc,
                        self.exc.__traceback__,
                    )
                )
        return self._fmt_exc

    def result(self) -> Any:
        if self.exc is not None:
            # TODO: add a test for the exception
            raise self.exc
        return self.ret



'''A function that builds a finite state machine of the nextline states.

The package "transitions" is used: https://github.com/pytransitions/transitions

State Diagram:

             .-------------.
             |   Created   |---.
             '-------------'   |
      initialize() |           |
                   |           |
                   V           |
             .-------------.   |
        .--->| Initialized |---.
reset() |    '-------------'   |
        |      |   | run()     |
        |------'   |           |
        |          v           |
        |    .-------------.   |
        |    |   Running   |   |
        |    '-------------'   |
        |          | finish()  |
        |          |           |
        |          V           |
        |    .-------------.   |  close()  .-------------.
        '----|  Finished   |-------------->|   Closed    |
             '-------------'               '-------------'

Example:

Define a model class with callbacks:

>>> class Model:
...     def on_enter_initialized(self, event):
...         print('enter the initialized state')
...
...     def on_exit_initialized(self, event):
...        print('exit the initialized state')
...
...     def on_reset(self, event):
...        print('resetting to the initialized state')
...        print(f'passed arguments: {event.args} {event.kwargs}')
...
...     def on_enter_running(self, event):
...        print('enter the running state')
...        self.finish()
...
...     def on_exit_running(self, event):
...        print('exit the running state')
...
...     def on_enter_finished(self, event):
...        print('enter the finished state')
...
...     def on_exit_finished(self, event):
...        print('exit the finished state')
...
...     def on_enter_closed(self, event):
...        print('enter the closed state')

Create a state machine with a model:

>>> model = Model()
>>> machine = build_state_machine(model=model, asyncio=False)

The model is in the "created" state.

>>> model.state
'created'

The method "initialize" triggers a transition.

>>> _ = model.initialize()
enter the initialized state

A callback is called and the state is now "initialized".

>>> model.state
'initialized'

Arguments to trigger methods are passed to callbacks.

>>> _ = model.reset(10, foo='bar')
resetting to the initialized state
passed arguments: (10,) {'foo': 'bar'}
exit the initialized state
enter the initialized state

A trigger method can be called from a callback.

>>> _ = model.run()
exit the initialized state
enter the running state
exit the running state
enter the finished state

Now the model is in the "finished" state.

>>> model.state
'finished'

Let's close the model.

>>> _ = model.close()
exit the finished state
enter the closed state

>>> model.state
'closed'

'''

from typing import Type

from transitions import Machine
from transitions.extensions import MachineFactory
from transitions.extensions.markup import MarkupMachine

CONFIG = {
    'name': 'nextline',
    'states': [
        'created',
        'initialized',
        'running',
        'finished',
        'closed',
    ],
    'transitions': [
        ['initialize', 'created', 'initialized'],
        ['run', 'initialized', 'running'],
        ['finish', 'running', 'finished'],
        ['close', ['created', 'initialized', 'finished'], 'closed'],
        {
            'trigger': 'reset',
            'source': ['initialized', 'finished'],
            'dest': 'initialized',
            'before': 'on_reset',
        },
    ],
    'initial': 'created',
    'queued': True,
    'send_event': True,
    # 'ignore_invalid_triggers': True,
}


def build_state_machine(model=None, graph=False, asyncio=True, markup=False) -> Machine:
    MachineClass: Type[Machine]
    if markup:
        MachineClass = MarkupMachine
    else:
        MachineClass = MachineFactory.get_predefined(graph=graph, asyncio=asyncio)

    machine = MachineClass(model=model, **CONFIG)  # type: ignore
    return machine



from __future__ import annotations

import sys
from contextlib import contextmanager
from typing import Any, Callable, TextIO


@contextmanager
def peek_stdout(callback: Callable[[str], Any]):
    textio = sys.stdout
    with peek_textio(textio, callback) as t:
        yield t


@contextmanager
def peek_stderr(callback: Callable[[str], Any]):
    textio = sys.stderr
    with peek_textio(textio, callback) as t:
        yield t


@contextmanager
def peek_textio(textio: TextIO, callback: Callable[[str], Any]):
    org_write = textio.write

    def write(s: str, /) -> int:
        callback(s)
        return org_write(s)

    textio.write = write  # type: ignore

    try:
        yield write
    finally:
        textio.write = org_write  # type: ignore



from __future__ import annotations

import time
from asyncio import Task, current_task
from typing import Any, Callable, List, Optional, Set

from nextline.utils.func import to_thread


class TaskDoneCallback:
    """Call a function when each registered asyncio task ends

    Parameters
    ----------
    done : callable, optional
        A function with one arg. Each time a registered asyncio task ends, this
        function will be called with the task object as the arg. The return
        value will be ignored.

        The `done` is optional. This class can be still useful to wait for all
        registered tasks to end.
    """

    def __init__(self, done: Optional[Callable[[Task], Any]] = None):
        self._done = done
        self._active: Set[Task] = set()
        self._exceptions: List[BaseException] = []

    def register(self, task: Optional[Task] = None) -> Task:
        """Add the current task by default, or the given task

        The callback function `done`, given at the initialization,
        will be called with the task object when the task ends.
        """
        if task is None:
            task = current_task()
            if task is None:
                raise RuntimeError("The current task not found")
        if task not in self._active:
            task.add_done_callback(self._callback)
            self._active.add(task)
        return task

    def close(self, interval: float = 0.001) -> None:
        """To be optionally called after all tasks are registered

        This method returns after all registered tasks end.

        The method cannot be called from a registered task.

        If exceptions are raised in the callback function, this method
        re-raises the first exception.

        """
        try:
            task = current_task()
        except RuntimeError:
            task = None
        if task is not None:
            if task in self._active:
                raise RuntimeError(
                    "The close() cannot be called from a registered task"
                )
        self._close(interval)
        self.reraise()

    async def aclose(self, interval: float = 0.001) -> None:
        """Awaitable version of close()"""
        task = current_task()
        if task is not None:
            if task in self._active:
                raise RuntimeError(
                    "The aclose() cannot be called from a registered task"
                )
        await to_thread(self._close, interval)
        self.reraise()

    def reraise(self) -> None:
        """Reraise the first exception occurred in the callback function"""
        if self._exceptions:
            raise self._exceptions[0]

    def _close(self, interval: float) -> None:
        while self._active:
            time.sleep(interval)

    def _callback(self, task: Task, *_, **__):
        """This method is given to asyncio.Task.add_done_callback()

        When called, this method, in turn, calls the callback function
        `done`, given at the initialization.

        Note: This method will be called by asyncio.Handle with the
        task object as the only argument. However, the method is
        accepting additional arbitrary arguments because an exception
        caused by an argument mismatch will be difficult to find; it
        will be caught by the asyncio.Handle and given to the
        exception handler of the event loop.
        """

        try:
            self._active.remove(task)
            if self._done:
                self._done(task)
        except BaseException as e:
            self._exceptions.append(e)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        self.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.aclose()



from __future__ import annotations

import asyncio
import linecache
from logging import getLogger
from typing import Any, AsyncIterator, Optional, Tuple

from .context import Context
from .fsm import Machine
from .types import PromptInfo, RunInfo, StdoutInfo, TraceInfo
from .utils import merge_aiters


class Nextline:
    """Nextline allows line-by-line execution of concurrent Python scripts

    Nextline supports concurrency with threading and asyncio. It uses multiple
    instances of Pdb, one for each thread and async task.


    Parameters
    ----------
    statement : str
        A Python code as a string
    run_no_start_from : int, optional
        The first run number. The default is 1.
    timeout_on_exit : float, optional
        The timeout in seconds to wait for the nextline to exit from the "with"
        block. The default is 3.

    """

    def __init__(
        self,
        statement: str,
        run_no_start_from: int = 1,
        timeout_on_exit: float = 3,
    ):

        logger = getLogger(__name__)
        logger.debug(f"statement starts with {statement[:25]!r}")
        logger.debug(f"The next run number will be {run_no_start_from}")

        self._context = Context(
            run_no_start_from=run_no_start_from,
            statement=statement,
        )
        self._timeout_on_exit = timeout_on_exit
        self._registry = self._context.registry

        self._started = False
        self._closed = False

        # TODO: instantiate Machine here without the running asyncio loop

    async def start(self) -> None:
        if self._started:
            return
        self._started = True
        await self._context.start()
        self._machine = Machine(self._context)
        await self._machine.initialize()  # type: ignore

    def __repr__(self):
        # e.g., "<Nextline 'running'>"
        return f"<{self.__class__.__name__} {self.state!r}>"

    async def close(self) -> None:
        """End gracefully"""
        if self._closed:
            return
        self._closed = True
        await self._machine.close()  # type: ignore
        await self._context.shutdown()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await asyncio.wait_for(self.close(), timeout=self._timeout_on_exit)

    async def run(self) -> None:
        """Execute the script and wait until it exits"""
        await self._machine.run()  # type: ignore
        await self._machine.finish()  # type: ignore

    def send_pdb_command(self, command: str, prompt_no: int, trace_no: int) -> None:
        logger = getLogger(__name__)
        logger.debug(f'send_pdb_command({command!r}, {prompt_no!r}, {trace_no!r})')
        self._machine.send_pdb_command(command, prompt_no, trace_no)

    def interrupt(self) -> None:
        self._machine.interrupt()

    def terminate(self) -> None:
        self._machine.terminate()

    def kill(self) -> None:
        self._machine.kill()

    def exception(self) -> Optional[BaseException]:
        """Uncaught exception from the last run"""
        return self._machine.exception()

    def result(self) -> Any:
        """Return value of the last run. always None"""
        return self._machine.result()

    async def reset(
        self,
        statement: Optional[str] = None,
        run_no_start_from: Optional[int] = None,
    ) -> None:
        """Prepare for the next run"""
        await self._machine.reset(  # type: ignore
            statement=statement,
            run_no_start_from=run_no_start_from,
        )

    @property
    def statement(self) -> str:
        """The script"""
        return self.get("statement")

    @property
    def state(self) -> Optional[str]:
        """The current condition of the script execution.

        The possible values are "initialized", "running", "finished",
        "closed"
        """
        try:
            return self._machine.state  # type: ignore
        except AttributeError:
            return None

    def subscribe_state(self) -> AsyncIterator[str]:
        return self.subscribe("state_name")

    @property
    def run_no(self):
        """The current run number"""
        return self.get("run_no")

    def subscribe_run_no(self) -> AsyncIterator[int]:
        return self.subscribe("run_no")

    def subscribe_trace_ids(self) -> AsyncIterator[Tuple[int]]:
        return self.subscribe("trace_nos")

    def get_source(self, file_name=None):
        if not file_name or file_name == self._registry.latest("script_file_name"):
            return self.get("statement").split("\n")
        return [e.rstrip() for e in linecache.getlines(file_name)]

    def get_source_line(self, line_no, file_name=None):
        """
        based on linecache.getline()
        https://github.com/python/cpython/blob/v3.9.5/Lib/linecache.py#L26
        """
        lines = self.get_source(file_name)
        if 1 <= line_no <= len(lines):
            return lines[line_no - 1]
        return ""

    def subscribe_run_info(self) -> AsyncIterator[RunInfo]:
        return self.subscribe("run_info")

    def subscribe_trace_info(self) -> AsyncIterator[TraceInfo]:
        return self.subscribe("trace_info")

    def subscribe_prompt_info(self) -> AsyncIterator[PromptInfo]:
        return self.subscribe("prompt_info")

    def subscribe_prompt_info_for(self, trace_no: int) -> AsyncIterator[PromptInfo]:
        return self.subscribe(f"prompt_info_{trace_no}")

        # Alternative implementation under development
        # return self._subscribe_prompt_info_for(trace_no)

    async def _subscribe_prompt_info_for(
        self, trace_no: int
    ) -> AsyncIterator[PromptInfo]:

        merged = merge_aiters(
            self.subscribe_prompt_info(),
            self.subscribe_trace_info(),
        )
        async for _, info in merged:
            if not info.trace_no == trace_no:  # type: ignore
                continue
            if isinstance(info, TraceInfo):
                if info.trace_no == trace_no:
                    if info.state == "finished":
                        break
                continue
            assert isinstance(info, PromptInfo)
            yield info

    def get(self, key) -> Any:
        return self._registry.latest(key)

    def subscribe(self, key, last: Optional[bool] = True) -> AsyncIterator[Any]:
        return self._registry.subscribe(key, last=last)

    def subscribe_stdout(self) -> AsyncIterator[StdoutInfo]:
        return self.subscribe("stdout", last=False)



from __future__ import annotations

import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from logging import getLogger
from typing import Any, Optional

from tblib import pickling_support

from . import process
from .count import RunNoCounter
from .process import QueueCommands, QueueRegistry, RunArg, RunResult
from .registrar import Registrar
from .types import PromptNo, RunNo, TraceNo
from .utils import MultiprocessingLogging, PubSub, Running, run_in_process

pickling_support.install()

SCRIPT_FILE_NAME = "<string>"


def _call_all(*funcs) -> None:
    '''Execute callables and ignore return values.

    Used to call multiple initializers in ProcessPoolExecutor.
    '''
    for func in funcs:
        func()


class Resource:
    def __init__(self) -> None:
        self.registry = PubSub[Any, Any]()
        self.q_commands: QueueCommands | None = None
        self._mp_context = mp.get_context('spawn')
        self._q_registry: QueueRegistry = self._mp_context.Queue()
        self._mp_logging = MultiprocessingLogging(mp_context=self._mp_context)
        self.registrar = Registrar(self.registry, self._q_registry)

    async def run(self, run_arg: RunArg) -> Running[RunResult]:
        func = partial(process.main, run_arg)
        self.q_commands = self._mp_context.Queue()
        initializer = partial(
            _call_all,
            self._mp_logging.initializer,
            partial(process.set_queues, self.q_commands, self._q_registry),
        )
        executor_factory = partial(
            ProcessPoolExecutor,
            max_workers=1,
            mp_context=self._mp_context,
            initializer=initializer,
        )
        return await run_in_process(func, executor_factory)

    async def open(self):
        await self._mp_logging.open()
        await self.registrar.open()

    async def close(self):
        await self.registrar.close()
        await self._mp_logging.close()
        await self.registry.close()


class Context:
    def __init__(self, run_no_start_from: int, statement: str):
        self._resource = Resource()
        self.registry = self._resource.registry
        self._registrar = self._resource.registrar
        self._run_no_count = RunNoCounter(run_no_start_from)
        self._running: Optional[Running[RunResult]] = None
        self._run_arg = RunArg(
            run_no=RunNo(run_no_start_from - 1),
            statement=statement,
            filename=SCRIPT_FILE_NAME,
        )
        self._run_result: RunResult | None = None
        self._q_commands: QueueCommands | None = None

    async def start(self):
        await self._resource.open()
        await self._registrar.script_change(
            script=self._run_arg['statement'], filename=self._run_arg['filename']
        )

    async def state_change(self, state_name: str):
        await self._registrar.state_change(state_name)

    async def shutdown(self):
        await self._resource.close()

    async def initialize(self) -> None:
        self._run_arg['run_no'] = self._run_no_count()
        self._run_result = None
        await self._registrar.state_initialized(self._run_arg['run_no'])
        await self._registrar.run_initialized(self._run_arg['run_no'])

    async def reset(
        self,
        statement: Optional[str] = None,
        run_no_start_from: Optional[int] = None,
    ):
        if statement:
            self._run_arg['statement'] = statement
            await self._registrar.script_change(
                script=statement, filename=self._run_arg['filename']
            )
        if run_no_start_from is not None:
            self._run_no_count = RunNoCounter(run_no_start_from)

    async def run(self) -> Running:
        self._running = await self._resource.run(self._run_arg)
        self._q_commands = self._resource.q_commands
        assert self._q_commands
        await self._registrar.run_start()
        return self._running

    def send_pdb_command(self, command: str, prompt_no: int, trace_no: int) -> None:
        logger = getLogger(__name__)
        logger.debug(f'send_pdb_command({command!r}, {prompt_no!r}, {trace_no!r})')
        if self._q_commands:
            self._q_commands.put((command, PromptNo(prompt_no), TraceNo(trace_no)))

    def interrupt(self) -> None:
        if self._running:
            self._running.interrupt()

    def terminate(self) -> None:
        if self._running:
            self._running.terminate()

    def kill(self) -> None:
        if self._running:
            self._running.kill()

    async def finish(self) -> None:
        assert self._running
        ret = await self._running
        self._q_commands = None
        self._running = None

        self._run_result = ret.returned or RunResult(ret=None, exc=None)

        if ret.raised:
            logger = getLogger(__name__)
            logger.exception(ret.raised)

        await self._registrar.run_end(
            result=self._run_result.fmt_ret,
            exception=self._run_result.fmt_exc,
        )

    def result(self) -> Any:
        assert self._run_result
        return self._run_result.result()

    def exception(self) -> Optional[BaseException]:
        assert self._run_result
        return self._run_result.exc

    async def close(self):
        pass



__all__ = [
    'ThreadDoneCallback',
    'TaskDoneCallback',
    'ThreadTaskDoneCallback',
    'agen_with_wait',
    'current_task_or_thread',
    'merge_aiters',
    'to_thread',
    'MultiprocessingLogging',
    'match_any',
    'peek_stderr',
    'peek_stdout',
    'peek_textio',
    'profile_func',
    'PubSub',
    'PubSubItem',
    'ExecutorFactory',
    'Running',
    'run_in_process',
    'ExcThread',
    'ThreadTaskIdComposer',
]

from .done_callback import TaskDoneCallback, ThreadDoneCallback, ThreadTaskDoneCallback
from .func import agen_with_wait, current_task_or_thread, merge_aiters, to_thread
from .multiprocessing_logging import MultiprocessingLogging
from .path import match_any
from .peek import peek_stderr, peek_stdout, peek_textio
from .profile import profile_func
from .pubsub import PubSub, PubSubItem
from .run import ExecutorFactory, Running, run_in_process
from .thread_exception import ExcThread
from .thread_task_id import ThreadTaskIdComposer
